# example_contracts = {
#     'chainlink_feed': '0x31e0a88fecB6eC0a411DBe0e9E76391498296EE9',
#     'uniswap_v2_pool': '0x9928e4046d7c6513326ccea028cd3e7a91c7590a',
#     'uniswap_v3_pool': '0x88e6a0c2ddd26feeb64f039a2c41296fcb3f5640',
#     'curve_pool': '0x06cb22615ba53e60d67bf6c341a0fd5e718e1655',
#     'sushi_pool': '0x397ff1542f962076d0bfe58ea045ffa2d347aca0',
#     'balancer_v2_pool': '0xc35bdda2e93c401c6645e0d8a0b2c86906c51710',
# }

# for target_contract_type, address in example_contracts.items():
#     print(target_contract_type)
#     detected_contract_type = await async_get_contract_type(address)
#     print(detected_contract_type)
