from __future__ import annotations

# getPoolOwnership
# getMaxRedeem
# getMaxBorrow
# getRewardSpeedsByPool
# getRewardSpeedsByPools
# getUnclaimedRewardsByDistributors
# getRewardsDistributorsBySupplier
