from __future__ import annotations


registry = '0x00000000000c2e074ec69a0dfb2997ba6c7d2e1e'
base_registrar = (
    '0x57f1887a8bf19b14fc0df6fd9b2acc9af147ea85'  # the erc721 token
)
resolver = '0x4976fb03c32e5b8cfe2b6ccb31c09ba78ebaba41'
reverse_records = '0x3671ae578e63fdf66ad4f3e12cc0c0d71ac7510c'
controller = '0x283af0b28c62c092c9727f1ee09c02ca627eb7f5'


old_resolvers = [
    '0x226159d592E2b063810a10Ebf6dcbADA94Ed68b8',
    '0xdaaf96c344f63131acadd0ea35170e7892d3dfba',
]
