# from __future__ import annotations

# from ctc import evm
# from ctc import spec


# def get_sushiswap_factory(network: spec.NetworkReference) -> spec.Address:
#     network = evm.get_network_chain_id(network)
#     if network == 1:
#         return '0xc0aee478e3658e2610c5f7a4a2e1777ce9e4f2ac'
#     else:
#         raise Exception()
